/****************************************************************************
**
** Copyright (C) 2023 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Design Viewer of the Qt Toolkit.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
****************************************************************************/

#include "projectmanager.h"

#include <QBuffer>
#include <QDirIterator>
#include <QQmlEngine>
#include <QQuickItem>
#include <QRandomGenerator>
#include <QRegularExpression>
#include <QResource>
#include <QTemporaryDir>

ProjectManager::ProjectManager(QObject *parent)
    : QObject(parent)
{}

ProjectManager::~ProjectManager()
{
    qDebug() << "ProjectManager destroyed. Cleaning up resources.";
    cleanupResources();
}

void ProjectManager::cleanupResources()
{
    const uchar *data;
    if (m_projectData.size()) {
        qDebug() << "Unregistering the previous data from QRC system. Path: " << m_projectPath
                 << " Size: " << m_projectData.size() << " bytes.";
        data = reinterpret_cast<const uchar *>(m_projectData.data());
        if (!QResource::unregisterResource(data, m_projectPath)) {
            qCritical() << "Cannot unregister the previous resource data.";
        }
    }
}

QString ProjectManager::unpackProject(const QByteArray &project)
{
    cleanupResources();

    m_projectData = project;
    const uchar *data;
    data = reinterpret_cast<const uchar *>(m_projectData.data());
    qDebug() << "Registering resource data. Size: " << m_projectData.size();

    const QString resourcePath{"/" + QString::number(QRandomGenerator::global()->generate())};
    m_projectPath = resourcePath;

    if (!QDir(resourcePath).removeRecursively()) {
        qWarning() << "Could not remove resource path: " << resourcePath;
    }

    if (!QResource::registerResource(data, resourcePath)) {
        qCritical() << "Can not load the resource data.";
        return {};
    }

    return QString(":") + resourcePath;
}

QString ProjectManager::findFile(const QString &dir, const QString &filter)
{
    QDirIterator it(dir, {filter}, QDir::Files);
    return it.next();
}

QString ProjectManager::readQmlProjectFile(const QString &qmlProjectFilePath)
{
    QFile file(qmlProjectFilePath);
    if (!file.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open Qml Project file! " << file.fileName() + ": "
                    << file.errorString();
        return {};
    }

    return QString::fromUtf8(file.readAll());
}

QString ProjectManager::getMainQmlFile(const QString &projectPath,
                                       const QString &qmlProjectFileContent)
{
    const QRegularExpression mainFileRegExp("mainFile:\\s*\"(.*)\"");
    const QRegularExpressionMatch mainFileMatch = mainFileRegExp.match(qmlProjectFileContent);

    if (!mainFileMatch.hasMatch()) {
        qCritical() << "No main file found in " << qmlProjectFileContent;
        return {};
    }

    qDebug() << "Found main file: " << mainFileMatch.captured(1);
    return projectPath + "/" + mainFileMatch.captured(1);
}

QStringList ProjectManager::getImportPaths(const QString &projectPath,
                                           const QString &qmlProjectFileContent)
{
    const QRegularExpression importPathsRegExp("importPaths:\\s*\\[\\s*(.*)\\s*\\]");
    const QRegularExpressionMatch importPathsMatch = importPathsRegExp.match(qmlProjectFileContent);

    if (!importPathsMatch.hasMatch()) {
        qWarning() << "No import paths found in " << qmlProjectFileContent;
        return {};
    }

    QStringList importPaths;
    for (const QString &path : importPathsMatch.captured(1).split(",")) {
        QString cleanedPath = path.trimmed();
        cleanedPath = projectPath + "/" + cleanedPath.mid(1, cleanedPath.length() - 2);
        if (QFileInfo::exists(cleanedPath)) {
            importPaths.append(cleanedPath);
        }
    }

    return importPaths;
}

bool ProjectManager::isQt6Project(const QString &qmlProjectFileContent)
{
    const QRegularExpression qt6ProjectRegExp("qt6Project:\\s*true");
    const QRegularExpressionMatch qt6ProjectMatch = qt6ProjectRegExp.match(qmlProjectFileContent);
    return qt6ProjectMatch.hasMatch();
}

bool ProjectManager::runProject(const QByteArray &project, const bool autoScaleProject)
{
    m_autoScaleProject = autoScaleProject;

    const QString projectPath = unpackProject(project);
    qDebug() << "Project location: " << projectPath;

    const QString qmlProjectFile = findFile(projectPath, "*.qmlproject");
    if (qmlProjectFile.isEmpty()) {
        qCritical() << "No \"*.qmlproject\" found in \"" << projectPath << "\".";
        return false;
    }

    const QString qmlProjectFileContent = readQmlProjectFile(qmlProjectFile);
    if (qmlProjectFileContent.isEmpty())
        return false;

    const QString mainQmlFilePath = getMainQmlFile(projectPath, qmlProjectFileContent);
    if (mainQmlFilePath.isEmpty())
        return false;

    const QString qtquickcontrols2File = findFile(projectPath, "qtquickcontrols2.conf");
    if (!qtquickcontrols2File.isEmpty()) {
        qputenv("QT_QUICK_CONTROLS_CONF", qtquickcontrols2File.toLatin1());
    }

    const QStringList importPaths = getImportPaths(projectPath, qmlProjectFileContent);
    if (!isQt6Project(qmlProjectFileContent)) {
        qWarning() << "This is not a Qt6 project.\nQt5 projects might work, but they are not "
                      "officially supported.";
    }

    qDebug() << "Project is parsed successfully.";
    qDebug() << "Qml Project File: " << qmlProjectFile;
    qDebug() << "Main Qml File: " << mainQmlFilePath;
    qDebug() << "Import Paths: " << importPaths;
    qDebug() << "Qt Quick Controls 2 File: " << qtquickcontrols2File;

    QUrl mainQmlUrl = QUrl::fromUserInput(mainQmlFilePath);
    QFile file(mainQmlUrl.path());
    if (!file.open(QIODevice::ReadOnly)) {
        qCritical() << "Could not open mainQmlfile for reading! " << file.fileName() << ": "
                    << file.errorString();
        return false;
    }

    qDebug() << "Initializing the qmlEngine";
    m_qmlEngine.reset(new QQmlEngine);
    m_qmlEngine->clearComponentCache();

    qDebug("Adding import paths");
    for (const QString &importPath : importPaths) {
        qDebug() << "-- Import path: " << importPath;
        m_qmlEngine->addImportPath(importPath);
    }

    QObject::connect(m_qmlEngine.data(),
                     &QQmlEngine::warnings,
                     this,
                     [&](const QList<QQmlError> &warnings) {
                         for (const auto &warning : warnings) {
                             qWarning() << warning.toString();
                         }
                     });

    qDebug() << "Loading mainQmlUrl: " << mainQmlUrl.toString();
    m_qmlComponent.reset(new QQmlComponent(m_qmlEngine.data()));

    m_qmlComponent->loadUrl(mainQmlUrl);

    qDebug() << "Waiting for qmlComponent to load";
    while (m_qmlComponent->isLoading())
        QCoreApplication::processEvents();

    qDebug() << "Checking if m_qmlComponent is ready";
    if (!m_qmlComponent->isReady()) {
        qCritical() << "m_qmlComponent is not ready. Reason:" << m_qmlComponent->errorString();
        return false;
    }
    qDebug() << "Creating top level object";
    QObject *topLevel = m_qmlComponent->create();
    if (!topLevel && m_qmlComponent->isError()) {
        qCritical() << "Error while creating Qml m_qmlComponent:" << m_qmlComponent->errorString();
        return false;
    }

    qDebug() << "Setting up the quickWindow";
    m_quickWindow.reset(qobject_cast<QQuickWindow *>(topLevel));
    if (m_quickWindow) {
        qDebug() << "Running with incubator controller";
        m_qmlEngine->setIncubationController(m_quickWindow->incubationController());
    } else {
        qWarning() << "Top level object is not a QQuickWindow. Trying QQuickView...";

        QQuickItem *contentItem = qobject_cast<QQuickItem *>(topLevel);
        if (!contentItem) {
            qCritical() << "Top level object cannot be casted to QQuickItem. Aborting.";
            return false;
        }

        qDebug() << "Initializing QQuickView";
        QQuickView *view = new QQuickView(m_qmlEngine.data(), nullptr);
        m_quickWindow.reset(view);
        view->setContent(mainQmlUrl, m_qmlComponent.data(), contentItem);
        view->setResizeMode(QQuickView::SizeViewToRootObject);
        m_quickWindow->setBaseSize(QSize(contentItem->width(), contentItem->height()));
    }
    showAppWindow();
    return true;
}

void ProjectManager::orientateWindow(Qt::ScreenOrientation orientation)
{
    if (!m_autoScaleProject)
        return;

    QQuickItem *contentItem = m_quickWindow->contentItem();
    QQuickItem *childItem{contentItem->childItems().at(0)};

    const QRect screenGeometry = QGuiApplication::primaryScreen()->geometry();
    const QString orientationString = (orientation == Qt::PortraitOrientation ? "portrait"
                                                                              : "landscape");
    qDebug() << "Adapting orientation to" << orientationString << "mode. enum_val:" << orientation
             << ". Initial sizing:" << Qt::endl
             << "-- Screen size: " << screenGeometry.height() << " x " << screenGeometry.width()
             << Qt::endl
             << "-- Quick window size: " << m_quickWindow->height() << " x "
             << m_quickWindow->width() << Qt::endl
             << "-- Child size: " << childItem->height() << " x " << childItem->width() << Qt::endl
             << "-- Child pos: " << childItem->x() << ", " << childItem->y() << Qt::endl
             << "-- Child scale: " << childItem->scale();

    const QSizeF newContentSize = childItem->size().scaled(screenGeometry.size().toSizeF(),
                                                           Qt::AspectRatioMode::KeepAspectRatio);

    const qreal newScale = newContentSize.height() / childItem->size().height();
    const qreal newX = (childItem->width() - screenGeometry.width()) / -2.0f;
    const qreal newY = (childItem->height() - screenGeometry.height()) / -2.0f;

    childItem->setScale(newScale);
    childItem->setPosition(QPointF(newX, newY));

    qDebug() << "Calculated props: height-width-scale-posX-posY" << Qt::endl
             << "--" << newContentSize.height() << "-" << newContentSize.width() << "-" << newScale
             << "-" << newX << "-" << newY << Qt::endl
             << "Final Sizing: height-width-scale-posX-posY" << Qt::endl
             << "--" << childItem->height() << "-" << childItem->width() << "-"
             << childItem->scale() << "-" << childItem->x() << "-" << childItem->y();
}

void ProjectManager::showAppWindow()
{
    qDebug("Initializing and showing the QML app window");

    QScreen *screen = QGuiApplication::primaryScreen();
    connect(screen, &QScreen::orientationChanged, this, &ProjectManager::orientateWindow);
    orientateWindow(screen->orientation());

    connect(m_quickWindow.data(), &QQuickWindow::closing, this, [this, screen]() {
        qDebug() << "QML app window is closing";
        disconnect(screen, &QScreen::orientationChanged, this, &ProjectManager::orientateWindow);
        // this signal is connected to the lambda in the backend
        // which will reset the project manager
        emit closingProject();
    });

    m_quickWindow->setFlags(Qt::Window | Qt::WindowStaysOnTopHint);
    m_quickWindow->showMaximized();
}

void ProjectManager::stopProject()
{
    if (!m_quickWindow || !m_quickWindow->isVisible())
        return;

    qDebug("Stopping the QML app window");
    m_quickWindow->close();
}

bool ProjectManager::isRunning()
{
    return m_quickWindow && m_quickWindow->isVisible();
}
